unit DbEzDateEdit;

{********************************************************************

    The TFlexDBDateEdit component is Copyright 1996
    by Tim Borman, All Rights Reserved.

    Version 0.9 (beta)

    This component may be freely distributed. It may be used
    or abused in any way desired provided that the author's
    copyright is recognized and respected.
    The author will not be held responsible for any problems
    arising from the use or mis-use of this component including,
    but not limited to, premature balding, nuclear warfare, or
    setting fire to the cat.

    If you find any bugs, have any suggestions for improvements,
    or make improvements yourself please tell me about them.
    Tim Borman, 72633.722@Compuserve.Com .


This component is designed to make date entry easy. No more
requirements to enter dates as carefully crafted sets of numbers.
Just tell it the field order (MDY etc), which fields are required,
and how to format the result. Months can be entered as numbers
or as month names (needing only enough letters to identify the
month). Field separators can be anything except a letter or number.
Years default to the current century if only two digits are entered.
Some examples, all using the default properties:
"D 1 96" -> Dec 1, 1996
"1/1/2001" -> Jan 1, 2001
"8/13" -> Aug 13, 1996
"ju 1" -> Jun 1, 1996

Credit where credit is due department:
I don't know who wrote the original freeware "DateEdit" but that
served as the basis for this. Not much of the original code remains
but much of the functionality is the same (especially in the popup)
and he/she deserves thanks. 

Usage notes:
Don't assume that the Date property is valid if the component is
currently focused. The Date property is only set when 1) a date is
selected from the popup, 2) the componenent is exited, or 3) when
a value is set in Date or SetDateText().  You can force the Date
to be updated by calling SetDateText(Text) if needed.


Properties:
  AllowBlank
    Whether a blank (null) entry is allowed.
    Default: True;
  DateOrder
    The order of the MDY fields.
    Default: Windows/CurrentUser default.
  RequiredFields
    Which of the MDY fields are required. Non-required fields
    will be filled in with the current system date if not entered.
    Default: MD
  DateFormat
    Display format of date. See FormatDateTime function in Delphi.
    If blank, uses Delphi's ShortDateFormat.
    Default: Blank.
  TodayString
    A string that can be entered to fill in the current date.
    Default:'/'
  Date
    Runtime only. The entered date, obviously, as TDateTime.
Procedures
  SetDateText
    The string is checked just as text entered into the edit
    would be.
  Popup
    Pops up the little calendar, same as clicking the button.
Events
  OnGetText
    Called immediately before a text value is processed. You can
    use this to alter the text before processing.
  OnSetDate
    Called after the date is set.


History: See EzDateEdit.Pas

********************************************************************}

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, StdCtrls, EzDatePopup, Buttons, Dialogs, Mask, DB,
  DBTables, EzDateEdit;

type

  TDateButton = class( TBitBtn )
  private
  protected
     procedure Click; override;
  public
  published
    constructor Create(AOwner: TComponent); override;
  end;

{ TLocalDBEdit }
{This is an exact duplicate of TDBEdit. The FDataLink property
 is private in TDBEdit as are the DataChange and UpdateData methods,
 so I can't descend from that. I could just put the TDBEdit
 functionality straight into TDbEzDateEdit, but this way if
 TDBEdit is changed in a later Delphi I can just cut this
 out and paste the new one in to get all the changes.}
  TLocalDBEdit = class(TCustomMaskEdit)
  private
    FDataLink: TFieldDataLink;
    FCanvas: TControlCanvas;
    FAlignment: TAlignment;
    FFocused: Boolean;
    procedure DataChange(Sender: TObject);
    procedure EditingChange(Sender: TObject);
    function GetDataField: string;
    function GetDataSource: TDataSource;
    function GetField: TField;
    function GetReadOnly: Boolean;
    function GetTextMargins: TPoint;
    procedure SetDataField(const Value: string);
    procedure SetDataSource(Value: TDataSource);
    procedure SetFocused(Value: Boolean);
    procedure SetReadOnly(Value: Boolean);
    procedure UpdateData(Sender: TObject);
    procedure WMCut(var Message: TMessage); message WM_CUT;
    procedure WMPaste(var Message: TMessage); message WM_PASTE;
    procedure CMEnter(var Message: TCMEnter); message CM_ENTER;
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure CMGetDataLink(var Message: TMessage); message CM_GETDATALINK;
  protected
    procedure Change; override;
    function EditCanModify: Boolean; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure Notification(AComponent: TComponent;
      Operation: TOperation); override;
    procedure Reset; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Field: TField read GetField;
  published
    property AutoSelect;
    property AutoSize;
    property BorderStyle;
    property CharCase;
    property Color;
    property Ctl3D;
    property DataField: string read GetDataField write SetDataField;
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property MaxLength;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PasswordChar;
    property PopupMenu;
    property ReadOnly: Boolean read GetReadOnly write SetReadOnly default False;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
  end;

  TDbEzDateEdit = class( TLocalDBEdit )
  private
    fAllowBlank:Boolean;
    fButton: TDateButton;
    fDate: TDateTime;
    fDateFormat:String;
    fDateOrder:TEzDateOrder;
    fRequiredFields:TEzDateRequiredFields;
    fTodayString:String;
    fOnSetDate:TNotifyEvent;
    fOnGetText:TEzDateGetTextEvent;
  protected
    procedure WMSize( var Message: TWMSize ); message WM_SIZE;
    procedure SetDate( Value: TDateTime );
    procedure DoExit; override;
    procedure DataChange(Sender: TObject); dynamic;
    procedure UpdateData(Sender: TObject); dynamic;
  public
    constructor Create( AOwner: TComponent ); override;
    destructor Destroy; override;
    procedure CreateParams( var Params: TCreateParams ); override;
    property Date: TDateTime read fDate write SetDate;
    procedure SetDateText(Value:String); dynamic;
    procedure SetDateFormat(Value:String);
    procedure SetDateOrder(Value:TEzDateOrder);
    procedure Popup; dynamic;
  published
    property AllowBlank:Boolean read fAllowBlank write fAllowBlank default True;
    property DateFormat:String read fDateFormat write SetDateFormat;
    property DateOrder:TEzDateOrder read fDateOrder write SetDateOrder;
    property RequiredFields:TEzDateRequiredFields read fRequiredFields write fRequiredFields;
    property TodayString: String read fTodayString write fTodayString;
    property OnGetText:TEzDateGetTextEvent read fOnGetText write fOnGetText;
    property OnSetDate:TNotifyEvent read fOnSetDate write fOnSetDate;
  end;

procedure Register;

var
  frmCalendar: TfrmEzDatePopup;

implementation

{--- TDateButton ---}
constructor TDateButton.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csReplicatable];
end;

procedure TDateButton.Click;
var
  EditParent: TDbEzDateEdit;
begin
  EditParent := TDbEzDateEdit( Parent );
  EditParent.fDataLink.Edit;
  frmCalendar := TfrmEzDatePopup.Create( editParent );
  if frmCalendar.Execute(EditParent.fDate) = mrOK
  then EditParent.SetDate(EditParent.fDate);
  frmCalendar.Free;
  inherited Click;
  EditParent.SetFocus;
end;

{--- TDbEzDateEdit ---}

constructor TDbEzDateEdit.Create( AOwner: TComponent );
var
  c:array[0..1] of Char;
begin
  inherited Create( AOwner );
  FButton := TDateButton.Create( Self );
  FButton.Width := 17;
  FButton.Height := 17;
  FButton.Visible := TRUE;
  FButton.Parent := self;
  FButton.Caption := '...';
  fButton.TabStop := False;
  fAllowBlank := True;
  ControlStyle := ControlStyle - [csSetCaption];
  fDateFormat := '';
  C[0] := '0'; {default}
  GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_IDATE, c, 2);
  fDateOrder := TEzDateOrder(Ord(C[0])-Ord('0'));
  fRequiredFields := [fdr_Month, fdr_Day];
  fTodayString := '/';

  FDataLink.OnDataChange := DataChange;
  FDataLink.OnUpdateData := UpdateData;
end;

procedure TDbEzDateEdit.CreateParams( var Params: TCreateParams );
begin
  inherited CreateParams( Params );
  Params.Style := Params.Style or WS_CLIPCHILDREN;
end;

destructor TDbEzDateEdit.Destroy;
begin
  FButton.Free;
  inherited Destroy;
end;

procedure TDbEzDateEdit.WMSize( var Message: TWMSize );
var
  nWidth: integer;
begin
  FButton.Height := Height - 3;
  nWidth := Height - 3;
  if nWidth > Width * 0.15 then nWidth := Round(Width * 0.15);
  FButton.Width := nWidth;
  FButton.Left := Width - nWidth - 3;
  FButton.Refresh;
end;

procedure TDbEzDateEdit.SetDate( Value: TDateTime );
var
  S:String;
begin
  fDate := Value;
  Modified := TRUE;
  if fDate = 0
  then Text := ''
  else begin
    if fDateFormat = ''
    then S := ShortDateFormat
    else S := fDateFormat;
    Text := FormatDateTime( S, fDate );
    end; {else}
  if Assigned(fOnSetDate)
  then fOnSetDate(Self);
end;

procedure TDbEzDateEdit.SetDateFormat(Value:String);
begin
  fDateFormat := Value;
  SetDate(fDate);
end;

procedure TDbEzDateEdit.SetDateOrder(Value:TEzDateOrder);
begin
  fDateOrder := Value;
  SetDate(fDate);
end;

procedure TDbEzDateEdit.DoExit;
begin
  try
    SetDateText(Text);
  except
     SetFocus;
     raise;
  end;
  inherited DoExit;
end;

procedure TDbEzDateEdit.Popup;
begin
  fButton.Click;
end;

procedure TDbEzDateEdit.SetDateText(Value:String);
var
  StringList:TStringList;
  Field,I1,I2:Integer;
  S:String;
  Error:Boolean;
  ThisYear,ThisMonth,ThisDay:Word;
  Year,Month,Day:Word;
  FieldOrder:String[3];
begin
  if Assigned(fOnGetText) then fOnGetText(Self, Value);
  if Value = '' then begin
    if fAllowBlank then begin
      fDate := 0;
      Text := '';
      Exit;
      end {then}
    else raise Exception.Create('Blank Date not Allowed');
    end; {if}
  if Value = TodayString then begin
    fDate := SysUtils.Date;
    if fDateFormat = ''
    then S := ShortDateFormat
    else S := fDateFormat;
    Text := FormatDateTime( S, fDate );
    end {then}
  else begin
    DecodeDate(SysUtils.Date, ThisYear,ThisMonth,ThisDay);
    Value := UpperCase(Value);
    StringList := TStringList.Create;
    try
      I1 := 1; {first, parse the string}
      while (I1 <= Length(Value))
        and not (Value[I1] in ['0'..'9', 'A'..'Z'])
      do Inc(I1);
      while I1 <= Length(Value) do begin
        I2 := I1;
        while (I2 <= Length(Value))
          and (Value[I2] in ['0'..'9', 'A'..'Z'])
        do Inc(I2);
        StringList.Add(Copy(Value, I1, I2-I1));
        while (I2 <= Length(Value))
          and not (Value[I2] in ['0'..'9', 'A'..'Z'])
        do Inc(I2);
        I1 := I2;
        end; {while}
      case fDateOrder of
        fdo_MDY:FieldOrder := 'MDY';
        fdo_DMY:FieldOrder := 'DMY';
        fdo_YMD:FieldOrder := 'YMD';
        end; {case}
      for Field := 1 to Length(FieldOrder) do begin {now analyse it}
        if StringList.Count > 0
        then S := StringList[0]
        else S := '';
        Error := False;
        case FieldOrder[Field] of
          'M':begin
            if (S = '') or (S[1] in ['0'..'9']) then begin {numeric month}
              try
                if S = ''
                then Month := 0
                else Month := StrToInt(S);
              except
                Month := 0;
                Error := True
              end;
              if not (Month in [1..12]) then Month := 0;
              end {then}
            else begin {one or more letters in month}
              Month := 0;
              I1 := 1;
              S := Copy(S, 1, 3); {only 3 chars needed to ID month}
              Error := True;
              repeat
                if S = UpperCase(Copy(ShortMonthNames[I1], 1, Length(S))) then begin
                  Month := I1;
                  I1 := 13;
                  Error := False;
                  end {then}
                else Inc(I1);
              until I1 = 13;
              end; {else}
            if Month = 0 then begin
              if fdr_Month in fRequiredFields
              then Error := True
              else Month := ThisMonth;
              end {then}
            else if StringList.Count > 0 then StringList.Delete(0);
            if Error then begin
              if S = ''
              then raise Exception.Create('Invalid Date "'+Value+'"')
              else raise Exception.Create('Invalid Month "'+S+'"');
              end; {if}
            end;
          'Y':begin
            try
              if S = ''
              then Year := 0
              else Year := StrToInt(S);
            except
              Year := 0;
              Error := True
            end;
            {default to current century if anything entered}
            if (Year < 100) and (S <> '')
            then Year := Year + (ThisYear div 100 * 100);
            if Year = 0 then begin
              if fdr_Year in fRequiredFields
              then Error := True
              else Year := ThisYear;
              end {then}
            else if StringList.Count > 0 then StringList.Delete(0);
            if Error then begin
              if S = ''
              then raise Exception.Create('Invalid Date "'+Value+'"')
              else raise Exception.Create('Invalid Year "'+S+'"');
              end; {if}
            end;
          'D':begin
            try
              if S = ''
              then Day := 0
              else Day := StrToInt(S);
            except
              Day := 0;
              Error := True
            end;
            if not (Day in [1..31]) then Day := 0;
            if Day = 0 then begin
              if fdr_Day in fRequiredFields
              then Error := True
              else Day := ThisDay;
              end {then}
            else if StringList.Count > 0 then StringList.Delete(0);
            if Error then begin
              if S = ''
              then raise Exception.Create('Invalid Date "'+Value+'"')
              else raise Exception.Create('Invalid Day "'+S+'"');
              end; {if}
            end;
          end; {case}
        end; {for}
      try
        fDate := EncodeDate(Year, Month, Day);
        if fDateFormat = ''
        then S := ShortDateFormat
        else S := fDateFormat;
        Text := FormatDateTime( S, fDate );
      except
        raise Exception.Create('Invalid Date "'+Value+'"');
      end;
    finally
      StringList.Free;
      end;
    end; {else}
end;


procedure TDbEzDateEdit.DataChange(Sender: TObject);
begin
  if FDataLink.Field <> nil then begin
    if FAlignment <> FDataLink.Field.Alignment then begin
      EditText := '';  {forces update}
      FAlignment := FDataLink.Field.Alignment;
      end;
    EditMask := FDataLink.Field.EditMask;
    if FDataLink.Field.IsNull
    then Date := 0
    else Date := (FDataLink.Field as TDateTimeField).AsDateTime;
    end {then}
  else begin
    FAlignment := taLeftJustify;
    EditMask := '';
    MaxLength := 0;
    EditText := '';
    end; {else}
end;

procedure TDbEzDateEdit.UpdateData(Sender: TObject);
var
  D:TDateTime;
begin
  SetDateText(Text);
  if fDate = 0
  then FDataLink.Field.Clear
  else begin
    {leave the fractional part of the field (the time)
     intact if it is there}
    D := (FDataLink.Field as TDateTimeField).AsDateTime;
    D := D - Trunc(D) + Trunc(fDate);
    (FDataLink.Field as TDateTimeField).AsDateTime := D;
    end; {else}
end;

{-------------------------}

{ TLocalDBEdit }

constructor TLocalDBEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  inherited ReadOnly := True;
  ControlStyle := ControlStyle + [csReplicatable];
  FDataLink := TFieldDataLink.Create;
  FDataLink.Control := Self;
  FDataLink.OnDataChange := DataChange;
  FDataLink.OnEditingChange := EditingChange;
  FDataLink.OnUpdateData := UpdateData;
end;

destructor TLocalDBEdit.Destroy;
begin
  FDataLink.Free;
  FDataLink := nil;
  FCanvas.Free;
  inherited Destroy;
end;

procedure TLocalDBEdit.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (FDataLink <> nil) and
    (AComponent = DataSource) then DataSource := nil;
end;

procedure TLocalDBEdit.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited KeyDown(Key, Shift);
  if (Key = VK_DELETE) or ((Key = VK_INSERT) and (ssShift in Shift)) then
    FDataLink.Edit;
end;

procedure TLocalDBEdit.KeyPress(var Key: Char);
begin
  inherited KeyPress(Key);
  if (Key in [#32..#255]) and (FDataLink.Field <> nil) and
    not FDataLink.Field.IsValidChar(Key) then
  begin
    MessageBeep(0);
    Key := #0;
  end;
  case Key of
    ^H, ^V, ^X, #32..#255:
      FDataLink.Edit;
    #27:
      begin
        FDataLink.Reset;
        SelectAll;
        Key := #0;
      end;
  end;
end;

function TLocalDBEdit.EditCanModify: Boolean;
begin
  Result := FDataLink.Edit;
end;

procedure TLocalDBEdit.Reset;
begin
  FDataLink.Reset;
  SelectAll;
end;

procedure TLocalDBEdit.SetFocused(Value: Boolean);
begin
  if FFocused <> Value then
  begin
    FFocused := Value;
    if (FAlignment <> taLeftJustify) and not IsMasked then Invalidate;
    FDataLink.Reset;
  end;
end;

procedure TLocalDBEdit.Change;
begin
  FDataLink.Modified;
  inherited Change;
end;

function TLocalDBEdit.GetDataSource: TDataSource;
begin
  Result := FDataLink.DataSource;
end;

procedure TLocalDBEdit.SetDataSource(Value: TDataSource);
begin
  FDataLink.DataSource := Value;
  if Value <> nil then Value.FreeNotification(Self);
end;

function TLocalDBEdit.GetDataField: string;
begin
  Result := FDataLink.FieldName;
end;

procedure TLocalDBEdit.SetDataField(const Value: string);
begin
  FDataLink.FieldName := Value;
end;

function TLocalDBEdit.GetReadOnly: Boolean;
begin
  Result := FDataLink.ReadOnly;
end;

procedure TLocalDBEdit.SetReadOnly(Value: Boolean);
begin
  FDataLink.ReadOnly := Value;
end;

function TLocalDBEdit.GetField: TField;
begin
  Result := FDataLink.Field;
end;

procedure TLocalDBEdit.DataChange(Sender: TObject);
begin
  if FDataLink.Field <> nil then
  begin
    if FAlignment <> FDataLink.Field.Alignment then
    begin
      EditText := '';  {forces update}
      FAlignment := FDataLink.Field.Alignment;
    end;
    EditMask := FDataLink.Field.EditMask;
    if FDataLink.Field.DataType = ftString then
      MaxLength := FDataLink.Field.Size else
      MaxLength := 0;
    if FFocused and FDataLink.CanModify then
      Text := FDataLink.Field.Text
    else
      EditText := FDataLink.Field.DisplayText;
  end else
  begin
    FAlignment := taLeftJustify;
    EditMask := '';
    MaxLength := 0;
    if csDesigning in ComponentState then
      EditText := Name else
      EditText := '';
  end;
end;

procedure TLocalDBEdit.EditingChange(Sender: TObject);
begin
  inherited ReadOnly := not FDataLink.Editing;
end;

procedure TLocalDBEdit.UpdateData(Sender: TObject);
begin
  ValidateEdit;
  FDataLink.Field.Text := Text;
end;

procedure TLocalDBEdit.WMPaste(var Message: TMessage);
begin
  FDataLink.Edit;
  inherited;
end;

procedure TLocalDBEdit.WMCut(var Message: TMessage);
begin
  FDataLink.Edit;
  inherited;
end;

procedure TLocalDBEdit.CMEnter(var Message: TCMEnter);
begin
  SetFocused(True);
  inherited;
end;

procedure TLocalDBEdit.CMExit(var Message: TCMExit);
begin
  try
    FDataLink.UpdateRecord;
  except
    SelectAll;
    SetFocus;
    raise;
  end;
  SetFocused(False);
  CheckCursor;
  DoExit;
end;

procedure TLocalDBEdit.WMPaint(var Message: TWMPaint);
var
  Left: Integer;
  Margins: TPoint;
  R: TRect;
  DC: HDC;
  PS: TPaintStruct;
  S: string;
begin
  if ((FAlignment = taLeftJustify) or FFocused) and
    not (csPaintCopy in ControlState) then
  begin
    inherited;
    Exit;
  end;
{ Since edit controls do not handle justification unless multi-line (and
  then only poorly) we will draw right and center justify manually unless
  the edit has the focus. }
  if FCanvas = nil then
  begin
    FCanvas := TControlCanvas.Create;
    FCanvas.Control := Self;
  end;
  DC := Message.DC;
  if DC = 0 then DC := BeginPaint(Handle, PS);
  FCanvas.Handle := DC;
  try
    FCanvas.Font := Font;
    with FCanvas do
    begin
      R := ClientRect;
      if not (NewStyleControls and Ctl3D) and (BorderStyle = bsSingle) then
      begin
        Brush.Color := clWindowFrame;
        FrameRect(R);
        InflateRect(R, -1, -1);
      end;
      Brush.Color := Color;
      if (csPaintCopy in ControlState) and (FDataLink.Field <> nil) then
      begin
        S := FDataLink.Field.DisplayText;
        case CharCase of
          ecUpperCase: S := AnsiUpperCase(S);
          ecLowerCase: S := AnsiLowerCase(S);
        end;
      end else
        S := EditText;
      if PasswordChar <> #0 then FillChar(S[1], Length(S), PasswordChar);
      Margins := GetTextMargins;
      case FAlignment of
        taLeftJustify: Left := Margins.X;
        taRightJustify: Left := ClientWidth - TextWidth(S) - Margins.X - 1;
      else
        Left := (ClientWidth - TextWidth(S)) div 2;
      end;
      TextRect(R, Left, Margins.Y, S);
    end;
  finally
    FCanvas.Handle := 0;
    if Message.DC = 0 then EndPaint(Handle, PS);
  end;
end;

procedure TLocalDBEdit.CMGetDataLink(var Message: TMessage);
begin
  Message.Result := Integer(FDataLink);
end;

function TLocalDBEdit.GetTextMargins: TPoint;
var
  DC: HDC;
  SaveFont: HFont;
  I: Integer;
  SysMetrics, Metrics: TTextMetric;
begin
  if NewStyleControls then
  begin
    if BorderStyle = bsNone then I := 0 else
      if Ctl3D then I := 1 else I := 2;
    Result.X := SendMessage(Handle, EM_GETMARGINS, 0, 0) and $0000FFFF + I;
    Result.Y := I;
  end else
  begin
    if BorderStyle = bsNone then I := 0 else
    begin
      DC := GetDC(0);
      GetTextMetrics(DC, SysMetrics);
      SaveFont := SelectObject(DC, Font.Handle);
      GetTextMetrics(DC, Metrics);
      SelectObject(DC, SaveFont);
      ReleaseDC(0, DC);
      I := SysMetrics.tmHeight;
      if I > Metrics.tmHeight then I := Metrics.tmHeight;
      I := I div 4;
    end;
    Result.X := I;
    Result.Y := I;
  end;
end;


procedure Register;
begin
  RegisterComponents('Data Controls', [TDbEzDateEdit]);
end;

end.
